/*	nonce.c  --  read/write random nonces to netscape cookie file
**
**	20000510 VRS 	Started.
**	20000530 VRS 	Now completely independent of apache.
*/

#include <sys/types.h> 
#include <stdio.h>
#include <errno.h>    
#include <stdlib.h>
#include <string.h>  
#include <ctype.h>
#include <sys/stat.h>

#include "nonce.h"


/*	Given cookie or prefs file pathname (relative to user's $HOME directory),
**	allocate and return full pathname and file contents.
**	Return 0:Success, > 0:Failure.
**	The caller must free these two pointers.
**
**	If all goes well, and it doesn't already exist, a backup copy of
**	the file will be created.
*/
int 	read_file(char *cf, char **path, char **data)
{
	struct stat	stbuf;
	FILE		*fp;
	char		*cp;
	int 		rc = 0;

	*path = *data = NULL;

	if (!cf  ||  !strlen(cf))  {
		return 1;		/*  No file specified							*/
		}
	if (!(*path = calloc(1, strlen(getenv("HOME")) + strlen(cf) + 6)))  {
		return 2;		/*  Unable to alloc path for *cf, skipping		*/
		}
	strcpy(*path, getenv("HOME"));  strcat(*path, "/");  strcat(*path, cf);

	if (stat(*path, &stbuf))  {
		return 3;		/*  Skipping cookie file						*/
		}
	else if (!(*data = calloc(1, stbuf.st_size + 1)))  {
		rc = 4;			/*  Unable to alloc data for *path, skipping	*/
		}
	else if (!(fp = fopen(*path, "r")))  {
		rc = 5; 		/*  Unable to read *path						*/
		}
	else  {
		for (cp = *data; (*cp = fgetc(fp)) != EOF; cp++) ;
		*cp = '\0';
		(void) fclose(fp);
		rc = 0;
		}

	if (! rc)  {				/* try to create backup cookie file 	*/
		strcat(*path, ".bak");
		if (stat(*path, &stbuf)  &&  errno == ENOENT)  {
			if ((fp = fopen(*path, "w")))  {
				for (cp = *data; *cp; (void) fputc(*cp++, fp)) ;
				(void) fclose(fp);
				}
			}
		(*path)[strlen(*path)-4] = '\0';
		}

	return rc;
}


/*	For given uri, return pointer to host:port substring and its length.
**	(Don't have a pool for ap_parse_uri_components()).
*/
void 	hostinfo(char *uri, char **host, int *len)
{
	char	*cp;

	*host = uri;
	*len = 0;

	if (!uri  ||  !*uri)  return;

	if (!(cp = strpbrk(uri, ":/")))  {
		*len = strlen(uri);
		return;
		}
	if (*cp == '/')  {
		*len = cp - uri;
		return;
		}
	if (*(cp+1) != '/')  {
		*len = strlen(uri);
		if ((cp = strchr(cp+1, '/')))  *len = cp - uri;
		return;
		}

	for (cp+=1; *cp  &&  *cp == '/'; cp++) ;
	*host = cp;
	*len = (cp = strchr(cp, '/'))? cp - *host: strlen(*host);
		
	return;
}


/*	Parse "val" text string of form "(NS|MSIE)[Version]:(Cookie|Language|None)"
**	Result is stored in int cookie->ns_auths or ->ms_auths as a 2-byte code:
**		low byte (0x0F) = version, forced to range 1-15
**		highbyte (0xF0) = AUTH_COOK  |  AUTH_LANG
*/
void	parse_auth_method(Cookie *cookie, const char *val)
{
	char	*versp; 
	int 	vers = 0;
	int 	*browser = NULL;

	if (!cookie)  return;
	if (!val)  val = "";

	if (!strncasecmp(val, "NS", 2)  ||  !strncasecmp(val, "Nets", 4))  {
		browser = &cookie->ns_auths;
		}
	if (!strncasecmp(val, "MS", 2)  ||  !strncasecmp(val, "IE", 2))  {
		browser = &cookie->ms_auths;
		}
	if (!browser)  browser = &cookie->ns_auths; 	/*  Netscape default	*/

	if ((versp = strpbrk(val, "0123456789")))  vers = atoi(versp);
	vers = (vers <= 0)? 1: (vers > 0x0F)? 0x0F: vers;
	*browser |= vers;

	if (!strncasecmp(val, "none", 4))  return;
	if (strstr(val, ":no") || strstr(val, ":No") || strstr(val, ":NO"))  return;
	*browser |= (strstr(val, ":C") || strstr(val, ":c"))? AUTH_COOK: AUTH_LANG;

	return;
}


/*	Overwrite NS cookie file, minus any entries which match domain_part.
**	The domainlist is formatted as: str1\0str2\0...strN\0\0.
**	Write new entry for local domain at end.
**	Entry = domain_part + static_part + random-cookie
**	Return 1:Success, 0:Failure.
*/
int 	write_ns_cookie_file(char *domainlist, char *nonce,
							 char *cfpath, char *cfdata)
{
	FILE			*fp;
	char 			*dp = domainlist;
	char			*cp = cfdata;
	int 			skipline = 0;
	const char		*lifetime = "2147126400";

	if (!nonce   ||  !domainlist  ||
		!cfdata  ||  !cfpath  || ! (fp = fopen(cfpath, "w")) )  return 0;

	while (*cp)  {
		for (dp=domainlist; *dp; dp += strlen(dp)+1)  {
			if (strncmp(cp, dp, strlen(dp)) == 0)  skipline = 1;
			}

		for ( ; *cp  &&  *cp != '\n'; cp++)  {
			if (! skipline)  fputc(*cp, fp);
			}
		if (*cp)  {
			if (! skipline)  fputc(*cp, fp);
			cp++;
			}
		}

	for (dp=domainlist; *dp; dp += strlen(dp)+1)  {
		fprintf(fp, "%s\t%s\t%s\t%s\t%s\t%s\t%s\n", dp, "FALSE", "/", "TRUE",
					lifetime, COOKIE_TAG, nonce);
		fprintf(fp, "%s\t%s\t%s\t%s\t%s\t%s\t%s\n", dp, "FALSE", "/", "FALSE",
					lifetime, COOKIE_TAG, nonce);
		}

	(void) fclose(fp);

	return 1;
}


void	clear_keyval(KeyVal *kv)
{
	if (!kv)  return;

	kv->keyp = NULL;    kv->keylen = 0;
	kv->valp = NULL;    kv->vallen = 0;
	return;
}


/*	Parse string for:  user_pref("key", ["]val["]);
**	Return length-delimited strings through KeyVal struct.
**	Return 0 on parse success, > 0 on parse fail.
**	Expects 1 entry per line.
*/
int 	parse_keyval(char *str, KeyVal *kv)
{
	char	*cp;
	int 	need_user = 4, need_key = 3, need_val = 2, need_close = 1;
	int 	state = need_user;

	if (!kv)  return 0;
	clear_keyval(kv);

	for (cp=str; *cp  &&  *cp != '\n'; cp++)  {
		while (*cp  &&  *cp != '\n'  &&  isspace((int)*cp))  cp++;
		if (!*cp  ||  *cp == '\n')  break;

		if (state == need_user)  {
			if (strncasecmp("user_pref", cp, 9))  continue;
			cp += 9;
			while (*cp  &&  *cp != '\n'  &&  isspace((int)*cp))  cp++;
			if (!*cp  ||  *cp != '(')  continue;
			state--;
			continue;
			}
		if (state == need_key)  {
			if (*cp == '"')  cp++;
			kv->keyp = cp;
			while (*cp  &&  *cp != '\n'  &&  *cp != '"'  &&  *cp != ',')  cp++;
			if (!*cp  ||  *cp == '\n')  break;
			kv->keylen = cp - kv->keyp;
			while (*cp  &&  *cp != '\n'  &&  *cp != ',')  cp++;
			if (!*cp  ||  *cp == '\n')  break;
			state--;
			continue;
			}
		if (state == need_val)  {
			if (*cp == '"')  cp++;
			kv->valp = cp;
			while (*cp  &&  *cp != '\n'  &&  *cp != '"'  &&  *cp != ')')  cp++;
			if (!*cp  ||  *cp == '\n')  break;
			kv->vallen = cp - kv->valp;
			while (*cp  &&  *cp != '\n'  &&  *cp != ')')  cp++;
			if (!*cp  ||  *cp == '\n')  break;
			state--;
			continue;
			}
		if (state == need_close)  {
			if (*cp == ';')  {
				state--;
				break;
				}
			}
		}

	return state;
}


/*	Given length-delimited string containing an Accept-Language: value,
**	edit out any old "COOKIE_TAG-<nonce>" strings and edit in a new one.
**	Use nonce=NULL to simply delete any "COOKIE_TAG-<nonce>" strings.
**	Returns a newly allocated string which the caller must free.
*/
char	*edit_lang(char *str, int len, char *nonce)
{
	char 	*cp, *ep, *new;
	int 	taillen;
	int 	newlen = len + strlen(", " COOKIE_TAG "-") + 1;

	if (nonce  &&  *nonce)  newlen += strlen(nonce);
	if (!len  ||  !*str  || 
		(new = (char *) calloc(1, newlen)) == NULL)  return NULL;

	strncpy(new, str, len);

	while ((cp = strstr(new, COOKIE_TAG "-")))  {	/* delete "ksslproxy-"	*/
		for (ep=cp; *ep  &&  !isspace((int)*ep); ep++) ;
		if (isspace((int)*ep)) ep++;
		taillen = (*ep)? strlen(ep): 0;
		(void) memcpy(cp, ep, taillen);
		*(cp + taillen) = '\0';
		}
	for (ep = new + strlen(new) - 1; ep > new; ep--)  {
		if (*ep  &&  !isspace((int)*ep)  &&  *ep != ',')  break;
		}
	ep++;

	if (nonce  &&  *nonce)  {
		if (ep > new  &&  *ep != ',')  (void) strcat(new, ", ");
		(void) strcat(new, COOKIE_TAG "-");
		(void) strcat(new, nonce);
		}
	
	return new;
}


/*	Overwrite NS preferences.js file, editing entries of interest.
**	Entries are of form:  user_pref("key", "value");
**	for keys in:
**		network.proxy.http, network.proxy.http_port,
**		network.proxy.ssl,  network.proxy.ssl_port,
**		network.proxy.type,	intl.accept_languages
**	Saves original entries first time called.
**	Calling with port=0 restores original entries.
**	Return 1:Success, 0:Failure.
*/
int 	write_ns_prefs_file(int port, char *nonce, char *cfpath, char *cfdata)
{
	FILE			*fp;
	char 			*cp, *newlang = NULL;
	char			*hp;
	KeyVal			kv, kv_http, kv_ssl, kv_lang;
	static char 	*restore = NULL;

	if (!nonce  ||  !cfpath  ||  ! (fp = fopen(cfpath, "w")) )  {
		return 0;
		}

	clear_keyval(&kv_http);
	clear_keyval(&kv_ssl);
	clear_keyval(&kv_lang);

	cp = cfdata;
	while (cp  &&  *cp)  {
		int 	skipline = 0;

		if (!parse_keyval(cp, &kv))  {		/*  0 on success  */
			if (!strncmp("network.proxy.type",		kv.keyp, kv.keylen))  {
				skipline = 1;
				kv_lang.valp = kv.valp; 		kv_lang.vallen = kv.vallen;
				}
			if (!strncmp("intl.accept_languages",	kv.keyp, kv.keylen))  {
				skipline = 1;
				kv_lang.keyp = kv.valp; 		kv_lang.keylen = kv.vallen;
				}
			/*
			if (!strncmp("network.proxy.http",		kv.keyp, kv.keylen))  {
				skipline = 1;
				kv_http.keyp = kv.valp; 		kv_http.keylen = kv.vallen;
				}
			if (!strncmp("network.proxy.http_port",	kv.keyp, kv.keylen))  {
				skipline = 1;
				kv_http.valp = kv.valp; 		kv_http.vallen = kv.vallen;
				}
			*/
			if (!strncmp("network.proxy.ssl",		kv.keyp, kv.keylen))  {
				skipline = 1;
				kv_ssl.keyp = kv.valp;			kv_ssl.keylen = kv.vallen;
				}
			if (!strncmp("network.proxy.ssl_port",	kv.keyp, kv.keylen))  {
				skipline = 1;
				kv_ssl.valp = kv.valp;			kv_ssl.vallen = kv.vallen;
				}
			}

		hp = cp;
		for ( ; *cp  &&  *cp != '\n'; cp++)  {
			if (! skipline)  fputc(*cp, fp);
			}
		if (*cp)  {
			if (! skipline)  fputc(*cp, fp);
			cp++;
			}
		}
	
	if (port  &&  !restore)  {
		restore = backup_ns_entries(&kv_http, &kv_ssl, &kv_lang);
		}

	if (port)  {
		fprintf(fp, "user_pref(\"network.proxy.type\", 1);\n");
		/*
		fprintf(fp, "user_pref(\"network.proxy.http\", \"127.0.0.1\");\n");
		fprintf(fp, "user_pref(\"network.proxy.http_port\", %d);\n", port);
		*/
		fprintf(fp, "user_pref(\"network.proxy.ssl\", \"127.0.0.1\");\n");
		fprintf(fp, "user_pref(\"network.proxy.ssl_port\", %d);\n", port);
		if ((newlang = edit_lang(kv_lang.keyp,kv_lang.keylen, nonce)))  {
			fprintf(fp, "user_pref(\"intl.accept_languages\", \"%s\");\n",
						newlang);
			free(newlang);
			}
		else  {
			fprintf(fp, "user_pref(\"intl.accept_languages\", \"%s-%s\");\n", 
					COOKIE_TAG, nonce);
			}
		}
	else if (restore)  {
		fprintf(fp, "%s", restore);
		free(restore);
		restore = NULL;
		}

	(void) fclose(fp);

	return 1;
}


#define	MAX_KVLEN(kvp)	((kvp->keylen > kvp->vallen)? kvp->keylen: kvp->vallen)

/*	Given 6 NS prefs file entries stored in 3 KeyVal structs,
**	allocate space for 6 entries and write backup copies.
**	Returns allocated backup buffer (which should be freed by caller.
**	Returns NULL on failure.
*/
char 	*backup_ns_entries(KeyVal *kv_http, KeyVal *kv_ssl, KeyVal *kv_lang)
{
	int 	max2, max_kv = 0;
	char	*long_entry = "user_pref('network.proxy.http_port', '');0";
	char	*bak = NULL;

	max2 = MAX_KVLEN(kv_http);		if (max2 > max_kv)  max_kv = max2;
	max2 = MAX_KVLEN(kv_ssl);		if (max2 > max_kv)  max_kv = max2;
	max2 = MAX_KVLEN(kv_lang);		if (max2 > max_kv)  max_kv = max2;

	if ((bak = (char *) calloc(6, strlen(long_entry) + max_kv)) == NULL)  {
		return NULL;
		}

	if (kv_lang->vallen)  {
		(void) sprintf(&bak[strlen(bak)], "user_pref(\"%s\", %.*s);\n",
					"network.proxy.type", kv_lang->vallen, kv_lang->valp);
		}
	/*
	if (kv_http->keylen)  {
		(void) sprintf(&bak[strlen(bak)], "user_pref(\"%s\", \"%.*s\");\n",
					"network.proxy.http", kv_http->keylen, kv_http->keyp);
		}
	if (kv_http->vallen)  {
		(void) sprintf(&bak[strlen(bak)], "user_pref(\"%s\", %.*s);\n", 
					"network.proxy.http_port", kv_http->vallen, kv_http->valp);
		}
	*/
	if (kv_ssl->keylen)  {
		(void) sprintf(&bak[strlen(bak)], "user_pref(\"%s\", \"%.*s\");\n",
					"network.proxy.ssl", kv_ssl->keylen, kv_ssl->keyp);
		}
	if (kv_ssl->vallen)  {
		(void) sprintf(&bak[strlen(bak)], "user_pref(\"%s\", %.*s);\n", 
					"network.proxy.ssl_port", kv_ssl->vallen, kv_ssl->valp);
		}
	if (kv_lang->keylen)  {
		(void) sprintf(&bak[strlen(bak)], "user_pref(\"%s\", \"%.*s\");\n",
					"intl.accept_languages", kv_lang->keylen, kv_lang->keyp);
		}

	return bak;
}

